;; Generated package description from switchy-window.el  -*- no-byte-compile: t -*-
(define-package "switchy-window" "1.4.0.20260208.102948" "A most-recently-used window switcher" '((emacs "25.1") (compat "29.1.4.1")) :commit "3ac4e522231a052e9beaa2d25478090d1215c7ae" :authors '(("Tassilo Horn" . "tsdh@gnu.org")) :maintainer '("Tassilo Horn" . "tsdh@gnu.org") :keywords '("windows") :url "https://sr.ht/~tsdh/switchy-window/")
